﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.IO;

namespace caldotnetc
{
    class Compiler
    {
        public void Compile(string fileToCompile)
        {
            Assembly asm = Assembly.GetExecutingAssembly();

            // TODO: errors
            Parser parser = new Parser(new BinaryReader(asm.GetManifestResourceStream("caldotnetc.Resources.CALDotNet.cgt")));

            if (!parser.Parse(new StreamReader(fileToCompile)))
            {
                Console.WriteLine("{0}:{1} {2}", parser.LineNumber, parser.LinePosition, parser.ErrorString);
                return;
            }

            CALRuntimeParams prms = new CALRuntimeParams();

            caldotnetc.Properties.Settings settings = caldotnetc.Properties.Settings.Default;

            prms.DBWrapperModuleName = settings.WrapperDll;
            prms.DBWrapperClassName = settings.WrapperClass;
            prms.DBWrapperSettings = settings.ConnectionString;

            Emitter em = new Emitter(Path.GetFileNameWithoutExtension(fileToCompile), 
                System.IO.Directory.GetCurrentDirectory() + "\\" + 
                Path.ChangeExtension(Path.GetFileName(fileToCompile), "exe"));
            TranslatorRuntime truntime = new TranslatorRuntime(em, prms);
            Translator translator = new Translator(truntime, em);

            try
            {
                translator.Translate(parser.SyntaxTree);
            }
            catch (TranslatorException ex)
            {
                Console.WriteLine(ex.Message);
                return;
            }
        }
    }

    public struct CALRuntimeParams
    {
        public string DBWrapperModuleName;
        public string DBWrapperClassName;
        public string DBWrapperSettings;
    }
}
