﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

/*
 * Eilutes:
 * eilutese ieskoma pagal zodzio dali
 * &, |, (), <>
 * tuscia eilute zymima '', del aiskumo galima zodzius rasyti tarp kabuciu ('mes nugalesim')
 * 
 * =kazkas (rasti eilutes kur stulpelio reiksme yra 'kazkas')
 * <>=kazkas (rasti eilutes kur stulpelio reiksme nera 'kazkas')
 * 
 * Skaiciai:
 * =, <>, >=, <, >, <=, (), .., !..
 * !5..6 (rasti eilutes kur stulpelio reiksme nera tarp 5 ir 6)
 * 
 * Bool: 
 * true | false
 *
 * Data
 * data galima ivesti nepilna, pvz 1108 bus pakeista i 2005-11-08
 * metus, menesi, diena atskirti galima bet kokiu simboliu, pvz 2005.07.05, 2005-06-04, 20050805, 050801, 5
 * =, <>, >, >=, <, >, <=, (), .., !.., simbolis d
 */

namespace caldotnet
{
    static internal class ConvertFilter
    {
        private static string not = "";
        private static string field = "";
        private static bool parsingString = false;
        private static bool equ = false;
        private static bool hadOperator = false;

        #region Strings

        private static string convertStrDoConvert(string symbol)
        {
            if (parsingString)
            {
                parsingString = false;
                not = "";

                if (equ)
                {
                    equ = false;
                    return "' " + symbol;
                }
                else
                    return "%') " + symbol;
            }
            else
                return symbol;
        }

        private static string convertStrMatch(Match m)
        {
            string val = m.Value.Trim();

            if ((val[0] == '\'') && (val[val.Length - 1] == '\''))
                val = val.Substring(1, val.Length - 2);

            if (val == string.Empty)
                equ = true;

            switch (val)
            {
                case "&": return convertStrDoConvert(" and ");
                case "|": return convertStrDoConvert(" or ");
                case "(": return convertStrDoConvert("(");
                case ")": return convertStrDoConvert(")");
                case "=":
                    equ = true;
                    return "";
                case "<>":
                    not = " not ";
                    return "";

                default:
                    if (!parsingString)
                    {
                        parsingString = true;

                        if (equ)
                            if (not != "")
                                return "[" + field + "] <> '" + val;
                            else
                                return "[" + field + "] = '" + val;

                        return "[" + field + "] " + not + "LIKE('%" + val;
                    }

                    return m.Value;
            }
        }
        #endregion

        #region Numbers

        private static string convertNumMatch(Match m)
        {
            if (m.Groups["operator"].Value != string.Empty)
            {
                hadOperator = true;
                return "[" + field + "]" + m.Value;
            }

            if (m.Groups["dots"].Value == "..")
            {
                string not = m.Groups["negation"].Value == string.Empty ? "" : "not";

                hadOperator = false;

                return "[" + field + "] " + not + " BETWEEN " + m.Groups["num1"].Value + " and " + m.Groups["num2"].Value + " ";
            }

            if (m.Groups["brackets"].Value != string.Empty)
            {
                hadOperator = false;

                return m.Value;
            }

            switch (m.Value)
            {
                case "|": return " or ";
                case "&": return " and ";
                default:
                    if (!hadOperator)
                        return "[" + field + "] =" + m.Value;
                    else
                    {
                        hadOperator = false;

                        return m.Value;
                    }
            }
        }
        #endregion

        #region Dates

        private static int convertDateGetNum(string str, ref int startPos, int maxNos)
        {
            int i = 0, pos = 0;

            while (true)
            {
                startPos--;

                if (startPos >= 0)
                {
                    if (str[startPos] < '0' || str[startPos] > '9')
                        startPos--;
                }
                else
                    return i;

                if (startPos >= 0)
                    i += (int)((int)(str[startPos] - '0') * Math.Pow(10, pos));

                if (++pos >= maxNos)
                    return i;
            }
        }

        private static string formatDateMatch(Match m)
        {
            int pos = m.Value.Length;

            int day = convertDateGetNum(m.Value, ref pos, 2);
            int month = convertDateGetNum(m.Value, ref pos, 2);
            int year = convertDateGetNum(m.Value, ref pos, 4);

            if (year == 0)
                year = DateTime.Now.Year;

            if (year < 100)
                year += 2000;

            if (month == 0)
                month = DateTime.Now.Month;

            DateTime dt = new DateTime(year, month, day);

            if ((year < 1800) || (year > 2100))
                throw new Exception("Date entered in wrong format.");

            return dt.ToShortDateString();
        }

        public static string formatDate(string str)
        {
            str = Regex.Replace(str, @"\b[dD]\b", DateTime.Now.ToShortDateString());
            str = Regex.Replace(str, @"\d{1,4}(.?(\d{1,2}))?(.?(\d{1,2}))?", formatDateMatch);

            return str;
        }

        private static string toUniversalDate(string str)
        {
            DateTime dt = DateTime.Parse(str);

            return dt.Year.ToString() + dt.Month.ToString().PadLeft(2, '0') + dt.Day.ToString().PadLeft(2, '0');
        }

        private static string convertDateMatch(Match m)
        {
            if (m.Groups["operator"].Value != string.Empty)
            {
                hadOperator = true;
                return "[" + field + "]" + m.Value;
            }

            if (m.Groups["dots"].Value == "..")
            {
                string not = m.Groups["negation"].Value == string.Empty ? "" : "not";

                hadOperator = false;

                return "[" + field + "] " + not + " BETWEEN '" + toUniversalDate(m.Groups["date1"].Value) + "' and '" + toUniversalDate(m.Groups["date2"].Value) + "' ";
            }

            if (m.Groups["brackets"].Value != string.Empty)
            {
                hadOperator = false;

                return m.Value;
            }

            switch (m.Value)
            {
                case "|": return " or ";
                case "&": return " and ";
                default:
                    if (!hadOperator)
                        return "[" + field + "] = '" + toUniversalDate(m.Value) + "'";
                    else
                    {
                        hadOperator = false;

                        return "'" + toUniversalDate(m.Value) + "'";
                    }
            }
        }

        #endregion

        #region Main

        private static string getDateOrIntFormat(string patternName, string forPattern)
        {
            return string.Format(@"(?<brackets>\)|\()|(?<operator>=|<|>|<>|>=|<=)|\&|\||((?<negation>!)?" +
                @"((?<{0}1>{1})\s*(?<dots>\.\.)\s*(?<{0}2>{1}))+)|{1}", patternName, forPattern);
        }

        private static string getDatePattern()
        {
            string sample = DateTime.Now.ToShortDateString();

            return Regex.Replace(sample, "[0-9]{1}", @"\d");
        }

        public static string Convert(string filter, string fieldName, Type fieldType)
        {
            field = fieldName;
            parsingString = false;

            switch (fieldType.ToString())
            {
                case "System.String":
                    filter = Regex.Replace(filter.Trim(), @"\s*\({1}\s*|\s*\){1}\s*|\s*\|{1}\s*|\s*&{1}\s*|\s*<>\s*|\s*={1}\s*|'.*?'|\w+", convertStrMatch);
                    filter += convertStrDoConvert("");
                    return filter;
                case "System.Int32":
                    return Regex.Replace(filter, getDateOrIntFormat("num", @"\d+(\.\d+)?"), convertNumMatch);
                case "System.Double":
                    return Regex.Replace(filter, getDateOrIntFormat("num", @"\d+(\.\d+)?"), convertNumMatch);
                case "System.Boolean":
                    switch (filter.Trim().ToLower())
                    {
                        case "true": return "[" + fieldName + "] = 1";
                        case "false": return "[" + fieldName + "] = 0";
                        default: throw new Exception("This field can only have values 'true' or 'false'.");
                    }
                case "System.DateTime":
                    return Regex.Replace(filter, getDateOrIntFormat("date", getDatePattern()), convertDateMatch);

                default: throw new NotImplementedException();
            }
        }
        #endregion
    }
}
