﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Reflection;
using System.Data;

namespace caldotnet
{
    public class DBInfo
    {
        public TableInfo[] Tables;
    }

    public class TableInfo
    {
        public string TableName;
        public FieldInfo[] Fields;
        public string[] PrimaryKeyFields;
        public ForeignKey[] ForeignKeys;

        public FieldInfo GetField(string name)
        {
            for (int i = 0; i < Fields.Length; i++)
                if (Fields[i].Name == name)
                    return Fields[i];

            return null;
        }
    }

    public class FieldInfo
    {
        public string Name;
        public string DBMSTypeName;
        public Type CALType;
        public int Precision;
        public int Scale;
        public int MaxLen;
    }

    public class ForeignKey
    {
        public string ReferencedTable;
        public string[] SourceFields;
        public string[] ReferencedFields;
    }

    public class DBWrapper
    {
        public static Record GetRecordInstance(IDBWrapper wrapper, string tableName)
        {
            DBInfo dbInfo = wrapper.GetDBInfo();

            foreach (TableInfo ti in dbInfo.Tables)
                if (ti.TableName == tableName)
                    return new Record(wrapper, ti);

            throw new Exception("Table '" + tableName + "' does not exist in specified database.");
        }

        public static IDBWrapper Create(string assemblyName, string className, string dbsettings)
        {
            Assembly dbassembly;
            Type type;
            object instance;

            dbassembly = Assembly.LoadFile(Path.GetFullPath(assemblyName));
            type = dbassembly.GetType(className);
            //TOMAYBEDO: check if it implements interface
            instance = dbassembly.CreateInstance(className, false, BindingFlags.ExactBinding, null,
                new object[] { dbsettings }, null, null);

            return (IDBWrapper)instance;
        }
    }

    public interface IDBWrapper
    {
        DBInfo GetDBInfo();
        void Rollback();
        void Commit();
        DataTable Query(string query);
        DataTable Query(string query, params DBParam[] parameters);
        void QueryNoResult(string query);
        void QueryNoResult(string query, params DBParam[] parameters);
    }

    public struct DBParam
    {
        public string paramName;
        public object value;

        public DBParam(string paramName, object value)
        {
            this.paramName = paramName;
            this.value = value;
        }
    }
}
