﻿using System;
using System.Collections;
using System.Data;
using System.Data.SqlClient;

using caldotnet;

namespace sqlservercal
{
    public class DBWrapper : IDBWrapper
    {
        private SqlConnection conn;
        private string connectionString;
        private DBInfo dbInfo;

        public DBWrapper(string settings)
        {
            connectionString = settings;
            conn = new SqlConnection(connectionString);
            dbInfo = GetDBInfoInternal();
        }

        public DBInfo GetDBInfo()
        {
            return dbInfo;
        }

        public void Rollback()
        {
            //TODO:
        }

        public void Commit()
        {
            //TODO:
        }

        private DBInfo GetDBInfoInternal()
        {
            DBInfo dbInfo = new DBInfo();
            conn.Open();

            // Tables
            DataTable tables = Query("SELECT name, object_id FROM sys.tables");
            dbInfo.Tables = new TableInfo[tables.Rows.Count];

            // Tables
            for (int i = 0; i < dbInfo.Tables.Length; i++)
            {
                TableInfo tinfo = new TableInfo();
                tinfo.TableName = (string)tables.Rows[i]["name"];
                dbInfo.Tables[i] = tinfo;

                string table_id = tables.Rows[i]["object_id"].ToString();

                // Fields
                DataTable fields = Query(string.Format(
                    "SELECT c.[name], t.[name] AS [type], c.max_length, c.[precision], c.scale " +
                    "FROM sys.columns as c, sys.types as t WHERE c.object_id = {0} AND " +
                    "t.system_type_id = c.system_type_id AND t.user_type_id = c.user_type_id", table_id));

                tinfo.Fields = new FieldInfo[fields.Rows.Count];

                for (int k = 0; k < tinfo.Fields.Length; k++)
                {
                    FieldInfo finfo = new FieldInfo();
                    tinfo.Fields[k] = finfo;

                    // Field types
                    finfo.Name = (string)fields.Rows[k]["name"];
                    finfo.DBMSTypeName = (string)fields.Rows[k]["type"];
                    finfo.CALType = mapDBMSTypeToCALType(finfo.DBMSTypeName);
                    finfo.MaxLen = (short)fields.Rows[k]["max_length"];
                    finfo.Precision = (byte)fields.Rows[k]["precision"];
                    finfo.Scale = (byte)fields.Rows[k]["scale"];
                }

                // Primary key
                DataTable primKey = Query(string.Format(
                    "SELECT c.[name] FROM sys.columns AS c, sys.index_columns AS ic, sys.indexes as i " +
                    "WHERE (c.object_id = {0}) AND (ic.object_id = {0}) AND (i.object_id = {0}) AND " +
                    "(i.is_primary_key = 1) AND (c.column_id = ic.column_id) AND (ic.index_id = i.index_id)",
                    table_id));

                if (primKey.Rows.Count == 0)
                    throw new Exception("Tables without primary key are not supported.");

                tinfo.PrimaryKeyFields = new string[primKey.Rows.Count];

                for (int k = 0; k < primKey.Rows.Count; k++)
                    tinfo.PrimaryKeyFields[k] = (string)primKey.Rows[k]["name"];

                // Foreign keys
                DataTable foreignKeys = Query(string.Format(
                    "SELECT fk.object_id as key_id, fk.referenced_object_id, t.[name] AS ref_name " +
                    "FROM sys.foreign_keys as fk, sys.tables as t WHERE " +
                    "(fk.parent_object_id = {0}) AND (fk.[type] = 'F') AND " +
                    "(t.object_id = fk.referenced_object_id)", table_id));

                tinfo.ForeignKeys = new ForeignKey[foreignKeys.Rows.Count];

                for (int k = 0; k < foreignKeys.Rows.Count; k++)
                {
                    ForeignKey fk = new ForeignKey();

                    string foreign_key_id = foreignKeys.Rows[k]["key_id"].ToString();
                    string referenced_table_id = foreignKeys.Rows[k]["referenced_object_id"].ToString();

                    tinfo.ForeignKeys[k] = fk;
                    fk.ReferencedTable = foreignKeys.Rows[k]["ref_name"].ToString();

                    DataTable srcFields = Query(string.Format(
                        "SELECT c.[name] FROM sys.columns AS c, sys.foreign_key_columns AS fc " +
                        "WHERE (c.object_id = {0}) AND " +
                        "(fc.constraint_object_id = {1}) AND " +
                        "(c.column_id = fc.parent_column_id)", table_id, foreign_key_id));

                    DataTable destFields = Query(string.Format(
                        "SELECT c.[name] FROM sys.columns AS c, sys.foreign_key_columns AS fc " +
                        "WHERE (c.object_id = {0}) AND " +
                        "(fc.constraint_object_id = {1}) AND " +
                        "(c.column_id = fc.referenced_column_id)", referenced_table_id, foreign_key_id));

                    int fieldsCount = srcFields.Rows.Count;
                    fk.SourceFields = new string[fieldsCount];
                    fk.ReferencedFields = new string[fieldsCount];

                    for (int j = 0; j < fieldsCount; j++)
                    {
                        fk.SourceFields[j] = srcFields.Rows[j]["name"].ToString();
                        fk.ReferencedFields[j] = destFields.Rows[j]["name"].ToString();
                    }
                }
            }

            return dbInfo;
        }

        internal object QuerySingleResult(string query)
        {
            SqlCommand cmd = new SqlCommand(query, conn);
            return cmd.ExecuteScalar();
        }

        public DataTable Query(string query)
        {
            DataTable dt = new DataTable();
            SqlDataAdapter adapter = new SqlDataAdapter(query, conn);

            if (conn.State != ConnectionState.Open)
                conn.Open();

            adapter.Fill(dt);

            return dt;
        }

        public DataTable Query(string query, params DBParam[] parameters)
        {
            if (conn.State != ConnectionState.Open)
                conn.Open();

            DataTable dt = new DataTable();
            SqlCommand cmd = new SqlCommand(query, conn);

            for (int i = 0; i < parameters.Length; i++)
                cmd.Parameters.AddWithValue(parameters[i].paramName, parameters[i].value);

            SqlDataAdapter adapter = new SqlDataAdapter(cmd);
            adapter.Fill(dt);

            return dt;
        }

        public void QueryNoResult(string query)
        {
            if (conn.State != ConnectionState.Open)
                conn.Open();

            SqlCommand cmd = new SqlCommand(query, conn);

            cmd.ExecuteNonQuery();
        }

        public void QueryNoResult(string query, params DBParam[] parameters)
        {
            if (conn.State != ConnectionState.Open)
                conn.Open();

            SqlCommand cmd = new SqlCommand(query, conn);

            for (int i = 0; i < parameters.Length; i++)
                cmd.Parameters.AddWithValue(parameters[i].paramName, parameters[i].value);

            cmd.ExecuteNonQuery();
        }

        internal Type mapDBMSTypeToCALType(string dbmsType)
        {
            // TODO: image, binary
            switch (dbmsType)
            {
                case "int": return Types.Int;
                case "float": return Types.Double;
                //case "": return "option";
                case "tinyint": return Types.Char;
                case "ntext": return Types.String;
                case "nvarchar": return Types.String;
                case "nchar": return Types.String;
                case "bit": return Types.Bool;
                //case "datetime": return Types.Date;
                //case "": return "Time";
                default:
                    throw new Exception("Type not supported: " + dbmsType);
            }
        }
    }
}
